/*++

Copyright (c) 2004-2013 Alexander A. Telyatnikov (Alter)

Module Name:
    PciDumpc.cpp

Abstract:
    This is simple utility that builds full list of PCI devices and
    assigned resources (IO/Memory ranges, IRQ, etc)

Author:
    Alexander A. Telyatnikov (Alter)

Environment:
    User mode only

Notes:

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Revision History:

--*/

#include <windows.h>
#include <stdio.h>
#include "..\Driver\PCIDump.h"
#include "SvcManLib.h"
#include "PciIds.h"
#include "Privilege.h"

#define PROD_VER_MJ  0
#define PROD_VER_MN  2
#define PROD_VER_SUB  "a"

BOOLEAN g_dump_dev = FALSE;
BOOLEAN g_dump_raw = FALSE;
BOOLEAN g_dump_bios = FALSE;
BOOLEAN g_dump_irt = FALSE;

HANDLE h;

char*
OnUpdateVendorId(
    ULONG VendorID
    ) 
{
    for(int i=0; v_names[i].str; i++) {
        if(v_names[i].id == VendorID) {
            return v_names[i].str;
        }
    }
    return "";
} // end OnUpdateVendorId()

char*
OnUpdateDeviceId(
    ULONG VendorID,
    ULONG DeviceID
    ) 
{
    ULONG z = (VendorID << 16) | DeviceID;
	
    for(int i=0; d_names[i].str; i++) {
        if(d_names[i].id == z) {
            return d_names[i].str;
        }
    }
    return "";	
} //end OnUpdateDeviceId()

char*
OnUpdateClassId(
    ULONG BaseClass
    ) 
{
    for(int i=0; c_names[i].str; i++) {
        if(c_names[i].id == BaseClass) {
            return c_names[i].str;
        }
    }
    return "";	
} // end OnUpdateClassId()

void
LengthToStr(
    PUCHAR buf,
    ULONG len
    )
{
    if(len < 1024) {
        sprintf((PCHAR)buf, "%d bytes ", len);
    } else
    if(len < 1024*1024) {
        sprintf((PCHAR)buf, "%d Kb ", len / 1024);
    } else {
        sprintf((PCHAR)buf, "%d Mb ", len / (1024*1024));
    }
} // end LengthToStr()

void
PrintBaseAddress(
    PCI_COMMON_CONFIG* pciData,
    ULONG i
    )
{
    ULONG tmp;
    ULONG base;
    ULONG len;
    BOOLEAN IoType;
    UCHAR buf[16];

    if(!pciData->u.type0.BaseAddresses[i])
        return;

    tmp = pciData->u.type0.BaseAddresses[i];
    IoType = (tmp & PCI_ADDRESS_IO_SPACE) ? TRUE : FALSE;

    // scan for first set bit, that's the length & alignment
    len = IoType ? 0x04 : 0x10;
    while (!(tmp & len) && len) {
        len *= 2;
    }

    base = tmp & (IoType ? ~0x3 : ~0xF);

    LengthToStr(buf, len);

    printf("\tBase Address %d: %08x\t%s\n",
        i,
        base,
        //buf,
        IoType ? "I/O" : "Mem");

/*    tmp = pciData->u.type0.BaseAddresses[i] & PCI_ADDRESS_MEMORY_TYPE_MASK; \
    SetDlgItemText(IDC_BASEADDRMT##i, \
    (tmp) ? (tmp == 2 ? PCITXT("20 bit") : PCITXT("64 bit")) : PCITXT("32 bit")); \
    pBt = (CButton*)GetDlgItem(IDC_BASEADDR_CHK_PF##i); \
    pBt->SetCheck((CurDevInfo.u.type0.BaseAddresses[i] & PCI_ADDRESS_MEMORY_PREFETCHABLE) ? (h != ((HANDLE)-1)) : 0);*/
} // end PrintBaseAddress()

char*
get_INTx(
    UCHAR int_pin
    )
{
    switch(int_pin) {
    case 1:
        return " INT#A";
    case 2:
        return " INT#B";
    case 3:
        return " INT#C";
    case 4:
        return " INT#D";
    }
    return "";
} // end get_INTx()

VOID
PrintDevInfo(
    ULONG              BusId,
    ULONG              DevId,
    ULONG              DevFunc,
    PCI_COMMON_CONFIG* pciData
    )
{
    BOOLEAN mf;
    ULONG i;
    
    printf("\nBus %02x  Device %02x  Function %02x\n", BusId,  
                                                        DevId,  
                                                        DevFunc);
    printf("    Vendor: %04x\t%s\n", pciData->VendorID, OnUpdateVendorId(pciData->VendorID));
    printf("    Device: %04x\t%s\n", pciData->DeviceID, OnUpdateDeviceId(pciData->VendorID, pciData->DeviceID));
    printf("    Revision: %02x\n", pciData->RevisionID);

    printf("    Device class: %02x\t%s\n", pciData->BaseClass, OnUpdateClassId(pciData->BaseClass));
    printf("    Device subclass: %02x\n", pciData->SubClass);
//    printf("\tDevice sub-subclass: ??\n");

    mf = (pciData->HeaderType & PCI_MULTIFUNCTION) != 0;
    if(mf)
        printf("    Multifunction device\n");

    if((pciData->HeaderType & ~PCI_MULTIFUNCTION) == PCI_DEVICE_TYPE) {

        printf("    Device subvendor: %04x", pciData->u.type0.SubVendorID);
        printf("    Device subsystem: %04x\n", pciData->u.type0.SubSystemID);

        PrintBaseAddress(pciData, 0);
        PrintBaseAddress(pciData, 1);
        PrintBaseAddress(pciData, 2);
        PrintBaseAddress(pciData, 3);
        PrintBaseAddress(pciData, 4);
        PrintBaseAddress(pciData, 5);

        if(pciData->u.type0.ROMBaseAddress) {
            printf("    ROM base: %08x\t%s\n", pciData->u.type0.ROMBaseAddress,
                (pciData->u.type0.ROMBaseAddress & PCI_ROMADDRESS_ENABLED) ? "Enabled" : "Disabled");
        }

        printf("    Interrupt line: %02x\tInterrupt pin: %02x%s\tMin_Gnt: %02x\tMaxLat: %02x\n",
            pciData->u.type0.InterruptLine,
            pciData->u.type0.InterruptPin, get_INTx(pciData->u.type0.InterruptPin),
            pciData->u.type0.MinimumGrant,
            pciData->u.type0.MaximumLatency);
        printf("    Cache line size: %02x\tLatency timer: %02x\tHeader type: %02x\tBIST: %02x\n",
            pciData->CacheLineSize,
            pciData->LatencyTimer,
            pciData->HeaderType,
            pciData->BIST);
        printf("    I/O: %d\tMem: %d\tBusMast: %d\tSpecial: %d\tMemInv: %d\n",
            pciData->Command & PCI_ENABLE_IO_SPACE     ? 1 : 0,
            pciData->Command & PCI_ENABLE_MEMORY_SPACE ? 1 : 0,
            pciData->Command & PCI_ENABLE_BUS_MASTER   ? 1 : 0,
            pciData->Command & PCI_ENABLE_SPECIAL_CYCLES ? 1 : 0,
            pciData->Command & PCI_ENABLE_WRITE_AND_INVALIDATE ? 1 : 0);

    }

    if(g_dump_raw) {
        for(i=0; i<sizeof(PCI_COMMON_CONFIG); i++) {
            if(!(i%16)) {
                printf("\n\t");
            } else
            if(!(i%8)) {
                printf("- ");
            }
            printf("%2.2x ", ((PUCHAR)pciData)[i]);
        }
        printf("\n\n");
    }

} // end PrintDevInfo()

VOID
BuildDevList()
{
    ULONG BusId=0;
    ULONG DevId=0;
    ULONG DevFunc=0;

    ULONG returned;

    PciDump_GetHeader_USER_IN  buf_in;
    PciDump_GetHeader_USER_OUT buf_out;

    while(h != ((HANDLE)-1)) {

        buf_in.BusNumber = BusId;
        buf_in.DeviceNumber = 0;
        buf_in.FunctionNumber = 0;
        
        if(!DeviceIoControl(h,IOCTL_PciDump_GetHeader,
                            &buf_in,sizeof(buf_in),
                            &buf_out,sizeof(buf_out),
                            &returned,NULL)
                  ||
            buf_out.status == 0) {
            break;
        }
        
        for(DevId=0; DevId<PCI_MAX_DEVICES; DevId++) {
            
            buf_in.BusNumber = BusId;
            buf_in.DeviceNumber = DevId;
            buf_in.FunctionNumber = 0;

            DeviceIoControl(h,IOCTL_PciDump_GetHeader,
                                &buf_in,sizeof(buf_in),
                                &buf_out,sizeof(buf_out),
                                &returned,NULL);

            if(buf_out.status == 2)
                continue;

/*            buf_in.BusNumber = BusId;
            buf_in.DeviceNumber = DevId;
            buf_in.FunctionNumber = 1;

            DeviceIoControl(h,IOCTL_PciDump_GetHeader,
                                &buf_in,sizeof(buf_in),
                                &buf_out,sizeof(buf_out),
                                &returned,NULL);

            if(buf_out.status == 2)
                continue;*/

            for(DevFunc=0; DevFunc<PCI_MAX_FUNCTION; DevFunc++) {

                buf_in.BusNumber = BusId;
                buf_in.DeviceNumber = DevId;
                buf_in.FunctionNumber = DevFunc;

                DeviceIoControl(h,IOCTL_PciDump_GetHeader,
                                    &buf_in,sizeof(buf_in),
                                    &buf_out,sizeof(buf_out),
                                    &returned,NULL);

                if(buf_out.status == 2)
                    continue;

                // Dump PCI device
                PrintDevInfo(BusId, DevId, DevFunc, &buf_out.pciData);
            }
        }

        BusId++;
    }
      
} // end BuildDevList()

VOID
GetBiosVer()
{
    ULONG returned;

    PciDump_GetBios32_USER_OUT buf_out;

    if(h != ((HANDLE)-1)) {
        if(DeviceIoControl(h,IOCTL_PciDump_GetBios32,
                            NULL,0,
                            &buf_out,sizeof(buf_out),
                            &returned,NULL)) {
            printf("BIOS32 ver %x.%2.2x\n", buf_out.VerMj, buf_out.VerMn);
            printf("Entry point at physical address 0x%8.8x\n", buf_out.EntryPoint);
            printf("\n");
        }
    }
} // end GetBiosVer()

VOID
BuildIRT()
{
    ULONG returned;
    ULONG i;
    PPciDump_GetIrqRoutingTable_USER_OUT buf_out;

    buf_out = (PPciDump_GetIrqRoutingTable_USER_OUT)GlobalAlloc(GMEM_FIXED, 65*1024);
    if(!buf_out)
        return;

    if(h != ((HANDLE)-1)) {
        if(DeviceIoControl(h,IOCTL_PciDump_GetIrqRoutingTable,
                            NULL,0,
                            buf_out,65*1024,
                            &returned,NULL)) {
            PPCI_BIOS32_IRQ_ROUTING_TABLE_ENTRY IrqRoutingTable = &buf_out->IrqRoutingTable[0];

            printf("PCI IRQ Routing Table size: %d(%#x) items\n", buf_out->ItemCount, buf_out->ItemCount);
            printf("PCI IRQ bitmap %#x\n", buf_out->IrqBitMap);
            for(i=0; i<buf_out->ItemCount; i++) {
                printf("  Bus %02x  Device %02x",
                    IrqRoutingTable[i].BusId,
                    IrqRoutingTable[i].DeviceNumber >> 3
                    );
                if(IrqRoutingTable[i].SlotNumber) {
                    printf("  slot %d\n",
                        IrqRoutingTable[i].SlotNumber
                        );
                } else {
                    printf("  onboard device\n"
                        );
                }
                printf("    INT#A Link %d, mask %x\n",
                    IrqRoutingTable[i].INTA.Link,
                    IrqRoutingTable[i].INTA.IRQBitMap
                    );
                printf("    INT#B Link %d, mask %x\n",
                    IrqRoutingTable[i].INTB.Link,
                    IrqRoutingTable[i].INTB.IRQBitMap
                    );
                printf("    INT#C Link %d, mask %x\n",
                    IrqRoutingTable[i].INTC.Link,
                    IrqRoutingTable[i].INTC.IRQBitMap
                    );
                printf("    INT#D Link %d, mask %x\n",
                    IrqRoutingTable[i].INTD.Link,
                    IrqRoutingTable[i].INTD.IRQBitMap
                    );
            }
            printf("\n");
        }
    }

    GlobalFree(buf_out);
} // end BuildIRT()

void print_help()
{
    printf(
        "Usage:\n"
        "  pcidumpc [options]\n"
        "    or\n"
        "  pcidumpc [options] > text_file.txt\n"
        "Options:\n"
        "  -d, --dev    dump brief PCI device info (assumed if no options specified)\n"
        "  -r, --raw    dump raw PCI config space\n"
        "  -b, --bios   print PCI BIOS32 version\n"
        "  -i, --irt    print PCI IRQ routing table\n"
        "  -a, --all    all available information\n"
        );
} // end print_help()

void main (int argc, char* argv[]) {

    LONG l=0;
    PUCHAR buff=NULL;
    PUCHAR buff_j=NULL;
    PUCHAR buff_i=NULL;
    LONG i=0, j=0;
    ULONG k;

    printf("PciDumpc v%d.%d%s (c) by Alter\nHome site http://alter.org.ua\n",
        PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);

    if(!Privilege(SE_LOAD_DRIVER_NAME, TRUE)) {
        printf("Warning: Insufficient privileges");
    }

    k = NtServiceInstall(NT_PciDump_SVC_NAME, ".");
    
    k = NtServiceIsRunning(NT_PciDump_SVC_NAME);
    if(k == 1) {
        // Ok
    } else
    if(k == 0) {
        k = NtServiceStart(NT_PciDump_SVC_NAME);
    }

    for(i=1; i<argc; i++) {
        if(!strcmp(argv[i], "--all") ||
           !strcmp(argv[i], "-a")) {
            g_dump_raw = TRUE;
            g_dump_dev = TRUE;
            g_dump_bios = TRUE;
            g_dump_irt = TRUE;
        } else
        if(!strcmp(argv[i], "--raw") ||
           !strcmp(argv[i], "-r")) {
            g_dump_raw = TRUE;
            g_dump_dev = TRUE;
        } else
        if(!strcmp(argv[i], "--bios") ||
           !strcmp(argv[i], "-b")) {
            g_dump_bios = TRUE;
        } else
        if(!strcmp(argv[i], "--irt") ||
           !strcmp(argv[i], "-i")) {
            g_dump_irt = TRUE;
        } else
        if(!strcmp(argv[i], "--help") ||
           !strcmp(argv[i], "-h") ||
           !strcmp(argv[i], "-?") ||
           !strcmp(argv[i], "/?") ||
           !strcmp(argv[i], "/h")) {
            print_help();
            exit(0);
        } else {
            print_help();
            exit(-1);
        }
    }
    if(argc == 1) {
        g_dump_dev = TRUE;
    }

    printf("Loading pci.ids ...\n\n");
    load_pci_info();
    if(k >= 0) {
        h = CreateFile(NT_PciDump_USER_NAME, GENERIC_READ | GENERIC_WRITE,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);

        if(h && h != ((HANDLE)-1)) {
            if(g_dump_bios) {
                GetBiosVer();
            }
            if(g_dump_irt) {
                BuildIRT();
            }
            if(g_dump_dev) {
                BuildDevList();
            }
            CloseHandle(h);
            h = ((HANDLE)-1);
        }
    } else {
        printf("Error: Can't load or install PciPumpEx driver");
        exit(-1);
    }

    exit(0);
}
